@@ -21,7 +21,7 @@ use GDPR::IAB::TCFv2::BitUtils qw<is_set
 use GDPR::IAB::TCFv2::PublisherRestrictions;
 use GDPR::IAB::TCFv2::RangeSection;
 
-our $VERSION = "0.081";
+our $VERSION = "0.082";
 
 use constant {
     CONSENT_STRING_TCF2_SEPARATOR => '.',
@@ -120,22 +120,34 @@ sub Parse {
 
     croak 'invalid vendor list version' if $self->vendor_list_version == 0;
 
-    # parse consent
+    # TODO parse special feature opt in
 
-    my $legitimate_interest_start = $self->_parse_vendor_consents();
+    #_parse_bitfield()
 
-    # parse legitimate interest
+    # TODO parse purpose section
+    # TODO parse purpose consent
 
-    my $pub_restrict_start =
-      $self->_parse_vendor_legitimate_interests($legitimate_interest_start);
+    # TODO parse purpose legitimate interest
+
+    # parse vendor section
+    # parse vendor consent
+
+    my $legitimate_interest_offset = $self->_parse_vendor_consents();
+
+    # parse vendor legitimate interest
+
+    my $pub_restrict_offset =
+      $self->_parse_vendor_legitimate_interests($legitimate_interest_offset);
+
+    # parse publisher section
 
     # parse publisher restrictions from section core string
 
-    $self->_parse_publisher_restrictions($pub_restrict_start);
+    $self->_parse_publisher_restrictions($pub_restrict_offset);
 
-    # parse section disclosed vendors if available
+    # TODO parse section disclosed vendors if available
 
-    # parse section publisher_tc if available
+    # TODO parse section publisher_tc if available
 
     return $self;
 }
@@ -149,7 +161,7 @@ sub tc_string {
 sub version {
     my $self = shift;
 
-    return get_uint6( $self->{data}, VERSION_OFFSET );
+    return scalar( get_uint6( $self->{data}, VERSION_OFFSET ) );
 }
 
 sub created {
@@ -171,7 +183,7 @@ sub last_updated {
 sub _get_epoch {
     my ( $self, $offset ) = @_;
 
-    my $deciseconds = get_uint36( $self->{data}, $offset );
+    my $deciseconds = scalar( get_uint36( $self->{data}, $offset ) );
 
     return (
         ( $deciseconds / 10 ),
@@ -182,94 +194,118 @@ sub _get_epoch {
 sub cmp_id {
     my $self = shift;
 
-    return get_uint12( $self->{data}, CMP_ID_OFFSET );
+    return scalar( get_uint12( $self->{data}, CMP_ID_OFFSET ) );
 }
 
 sub cmp_version {
     my $self = shift;
 
-    return get_uint12( $self->{data}, CMP_VERSION_OFFSET );
+    return scalar( get_uint12( $self->{data}, CMP_VERSION_OFFSET ) );
 }
 
 sub consent_screen {
     my $self = shift;
 
-    return get_uint6( $self->{data}, CONSENT_SCREEN_OFFSET );
+    return scalar( get_uint6( $self->{data}, CONSENT_SCREEN_OFFSET ) );
 }
 
 sub consent_language {
     my $self = shift;
 
-    return get_char6_pair( $self->{data}, CONSENT_LANGUAGE_OFFSET );
+    return scalar( get_char6_pair( $self->{data}, CONSENT_LANGUAGE_OFFSET ) );
 }
 
 sub vendor_list_version {
     my $self = shift;
 
-    return get_uint12( $self->{data}, VENDOR_LIST_VERSION_OFFSET );
+    return scalar( get_uint12( $self->{data}, VENDOR_LIST_VERSION_OFFSET ) );
 }
 
 sub policy_version {
     my $self = shift;
 
-    return get_uint6( $self->{data}, POLICY_VERSION_OFFSET );
+    return scalar( get_uint6( $self->{data}, POLICY_VERSION_OFFSET ) );
 }
 
 sub is_service_specific {
     my $self = shift;
 
-    return is_set( $self->{data}, SERVICE_SPECIFIC_OFFSET );
+    return scalar( is_set( $self->{data}, SERVICE_SPECIFIC_OFFSET ) );
 }
 
 sub use_non_standard_stacks {
     my $self = shift;
 
-    return is_set( $self->{data}, USE_NON_STANDARD_STACKS_OFFSET );
+    return scalar( is_set( $self->{data}, USE_NON_STANDARD_STACKS_OFFSET ) );
 }
 
 sub is_special_feature_opt_in {
     my ( $self, $id ) = @_;
 
-    croak "invalid special feature id $id: must be between 1 and 12"
-      if $id < 1 || $id > 12;
+    croak
+      "invalid special feature id $id: must be between 1 and @{[ MAX_SPECIAL_FEATURE_ID ]}"
+      if $id < 1 || $id > MAX_SPECIAL_FEATURE_ID;
 
-    return is_set( $self->{data}, SPECIAL_FEATURE_OPT_IN_OFFSET + $id - 1 );
+    return $self->_safe_is_special_feature_opt_in($id);
+}
+
+sub _safe_is_special_feature_opt_in {
+    my ( $self, $id ) = @_;
+
+    return
+      scalar(
+        is_set( $self->{data}, SPECIAL_FEATURE_OPT_IN_OFFSET + $id - 1 ) );
 }
 
 sub is_purpose_consent_allowed {
     my ( $self, $id ) = @_;
 
-    croak "invalid purpose id $id: must be between 1 and 24"
-      if $id < 1 || $id > 24;
+    croak "invalid purpose id $id: must be between 1 and @{[ MAX_PURPOSE_ID ]}"
+      if $id < 1 || $id > MAX_PURPOSE_ID;
 
-    return is_set( $self->{data}, PURPOSE_CONSENT_ALLOWED_OFFSET + $id - 1 );
+    return $self->_safe_is_purpose_consent_allowed($id);
+}
+
+sub _safe_is_purpose_consent_allowed {
+    my ( $self, $id ) = @_;
+    return
+      scalar(
+        is_set( $self->{data}, PURPOSE_CONSENT_ALLOWED_OFFSET + $id - 1 ) );
 }
 
 sub is_purpose_legitimate_interest_allowed {
     my ( $self, $id ) = @_;
 
-    croak "invalid purpose id $id: must be between 1 and 24"
-      if $id < 1 || $id > 24;
+    croak "invalid purpose id $id: must be between 1 and @{[ MAX_PURPOSE_ID ]}"
+      if $id < 1 || $id > MAX_PURPOSE_ID;
 
-    return is_set( $self->{data}, PURPOSE_LIT_ALLOWED_OFFSET + $id - 1 );
+    return $self->_safe_is_purpose_legitimate_interest_allowed($id);
+}
+
+sub _safe_is_purpose_legitimate_interest_allowed {
+    my ( $self, $id ) = @_;
+
+    return
+      scalar( is_set( $self->{data}, PURPOSE_LIT_ALLOWED_OFFSET + $id - 1 ) );
 }
 
 sub purpose_one_treatment {
     my $self = shift;
 
-    return is_set( $self->{data}, PURPOSE_ONE_TREATMENT_OFFSET );
+    return scalar( is_set( $self->{data}, PURPOSE_ONE_TREATMENT_OFFSET ) );
 }
 
 sub publisher_country_code {
     my $self = shift;
 
-    return get_char6_pair( $self->{data}, PUBLISHER_COUNTRY_CODE_OFFSET );
+    return
+      scalar( get_char6_pair( $self->{data}, PUBLISHER_COUNTRY_CODE_OFFSET ) );
 }
 
 sub max_vendor_id_consent {
     my $self = shift;
 
-    return get_uint16( $self->{data}, MAX_VENDOR_ID_CONSENT_OFFSET );
+    return scalar( get_uint16( $self->{data}, MAX_VENDOR_ID_CONSENT_OFFSET ) );
 }
 
 sub max_vendor_id_legitimate_interest {
@@ -352,41 +388,40 @@ sub TO_JSON {
         publisher_country_code  => $self->publisher_country_code,
         special_features_opt_in => $self->_format_json_subsection(
             map {
-                [ $_ => $self->is_special_feature_opt_in($_) ? $true : $false ]
+                [     $_ => $self->_safe_is_special_feature_opt_in($_)
+                    ? $true
+                    : $false
+                ]
             } 1 .. MAX_SPECIAL_FEATURE_ID
         ),
         purpose => {
-        consents => $self->_format_json_subsection(
-            map {
-                [     $_ => $self->is_purpose_consent_allowed($_)
-                    ? $true
-                    : $false
-                ]
-            } 1 .. MAX_PURPOSE_ID,
-        ),
-        legitimate_interests => $self->_format_json_subsection(
-            map {
-                [     $_ => $self->is_purpose_legitimate_interest_allowed($_)
-                    ? $true
-                    : $false
-                ]
-            } 1 .. MAX_PURPOSE_ID,
-        ),
-        },
-        vendor => {
             consents => $self->_format_json_subsection(
-                map { [ $_ => $self->vendor_consent($_) ? $true : $false ] }
-                  1 .. $self->max_vendor_id_consent,
-            ),
-            legitimate_interests => $self->_format_json_subsection(
                 map {
-                    [     $_ => $self->vendor_legitimate_interest($_)
+                    [     $_ => $self->_safe_is_purpose_consent_allowed($_)
                         ? $true
                         : $false
                     ]
-                } 1 .. $self->max_vendor_id_legitimate_interest,
+                } 1 .. MAX_PURPOSE_ID,
             ),
-        }
+            legitimate_interests => $self->_format_json_subsection(
+                map {
+                    [   $_ =>
+                          $self->_safe_is_purpose_legitimate_interest_allowed(
+                            $_)
+                        ? $true
+                        : $false
+                    ]
+                } 1 .. MAX_PURPOSE_ID,
+            ),
+        },
+        vendor => {
+            consents             => $self->{vendor_consents}->TO_JSON,
+            legitimate_interests =>
+              $self->{vendor_legitimate_interests}->TO_JSON,
+        },
+        publisher => {
+            restrictions => $self->{publisher_restrictions}->TO_JSON,
+        },
     };
 }
 
@@ -394,17 +429,17 @@ sub TO_JSON {
 sub _parse_vendor_consents {
     my $self = shift;
 
-    my ( $vendor_consents, $legitimate_interest_start );
+    my ( $vendor_consents, $legitimate_interest_offset );
 
     if ( $self->_is_vendor_consent_range_encoding ) {
-        ( $vendor_consents, $legitimate_interest_start ) =
+        ( $vendor_consents, $legitimate_interest_offset ) =
           $self->_parse_range_section(
             $self->max_vendor_id_consent,
             VENDOR_CONSENT_OFFSET
           );
     }
     else {
-        ( $vendor_consents, $legitimate_interest_start ) =
+        ( $vendor_consents, $legitimate_interest_offset ) =
           $self->_parse_bitfield(
             $self->max_vendor_id_consent,
             VENDOR_CONSENT_OFFSET
@@ -413,73 +448,75 @@ sub _parse_vendor_consents {
 
     $self->{vendor_consents} = $vendor_consents;
 
-    return $legitimate_interest_start;
+    return $legitimate_interest_offset;
 }
 
 sub _parse_vendor_legitimate_interests {
-    my ( $self, $legitimate_interest_start ) = @_;
+    my ( $self, $legitimate_interest_offset ) = @_;
 
-    my $legitimate_interest_max_vendor =
-      get_uint16( $self->{data}, $legitimate_interest_start );
+    my ($legitimate_interest_max_vendor,
+        $is_vendor_legitimate_interest_range_offset
+    ) = get_uint16( $self->{data}, $legitimate_interest_offset );
 
     $self->{legitimate_interest_max_vendor} = $legitimate_interest_max_vendor;
 
     my $data_size = length( $self->{data} );
+
     croak
-      "invalid consent data: no legitimate interest start position (got $legitimate_interest_start + 16 but $data_size)"
-      if $legitimate_interest_start + 16 > $data_size;
+      "invalid consent data: no legitimate interest start position (got $is_vendor_legitimate_interest_range_offset but $data_size)"
+      if $is_vendor_legitimate_interest_range_offset > $data_size;
 
-    my $is_vendor_legitimate_interest_range =
-      is_set( $self->{data}, $legitimate_interest_start + 16 );
+    my ($is_vendor_legitimate_interest_range,
+        $vendor_legitimate_interests_offset
+    ) = is_set( $self->{data}, $is_vendor_legitimate_interest_range_offset );
 
-    my ( $vendor_legitimate_interests, $pub_restrict_start );
+    my ( $vendor_legitimate_interests, $pub_restrict_offset );
 
     if ($is_vendor_legitimate_interest_range) {
-        ( $vendor_legitimate_interests, $pub_restrict_start ) =
+        ( $vendor_legitimate_interests, $pub_restrict_offset ) =
           $self->_parse_range_section(
             $self->max_vendor_id_legitimate_interest,
-            $legitimate_interest_start + 17
+            $vendor_legitimate_interests_offset
           );
     }
     else {
-        ( $vendor_legitimate_interests, $pub_restrict_start ) =
+        ( $vendor_legitimate_interests, $pub_restrict_offset ) =
           $self->_parse_bitfield(
             $self->max_vendor_id_legitimate_interest,
-            $legitimate_interest_start + 17
+            $vendor_legitimate_interests_offset
           );
     }
 
     $self->{vendor_legitimate_interests} = $vendor_legitimate_interests;
 
-    return $pub_restrict_start;
+    return $pub_restrict_offset;
 }
 
 sub _parse_publisher_restrictions {
-    my ( $self, $pub_restrict_start ) = @_;
+    my ( $self, $pub_restrict_offset ) = @_;
 
-    my $num_restrictions = get_uint12( $self->{data}, $pub_restrict_start );
+    my ( $num_restrictions, $next_offset ) =
+      get_uint12( $self->{data}, $pub_restrict_offset );
 
     my %restrictions;
 
-    my $current_offset = $pub_restrict_start + 12;
-
     for ( 1 .. $num_restrictions ) {
-        my $purpose_id = get_uint6( $self->{data}, $current_offset );
-        $current_offset += 6;
-        my $restriction_type = get_uint2( $self->{data}, $current_offset );
-        $current_offset += 2;
+        my ( $purpose_id, $restriction_type, $vendor_restrictions );
 
-        my ( $vendor_restrictions, $next_offset ) =
-          $self->_parse_range_section(
+        ( $purpose_id, $next_offset ) =
+          get_uint6( $self->{data}, $next_offset );
+
+        ( $restriction_type, $next_offset ) =
+          get_uint2( $self->{data}, $next_offset );
+
+        ( $vendor_restrictions, $next_offset ) = $self->_parse_range_section(
             ASSUMED_MAX_VENDOR_ID,
-            $current_offset
-          );
+            $next_offset
+        );
 
         $restrictions{$purpose_id} ||= {};
 
         $restrictions{$purpose_id}->{$restriction_type} = $vendor_restrictions;
-
-        $current_offset = $next_offset;
     }
 
     my $publisher_restrictions = GDPR::IAB::TCFv2::PublisherRestrictions->new(
@@ -488,7 +525,7 @@ sub _parse_publisher_restrictions {
 
     $self->{publisher_restrictions} = $publisher_restrictions;
 
-    return $current_offset;
+    return $next_offset;
 }
 
 sub _get_core_tc_string {
@@ -529,31 +566,35 @@ sub _decode_base64url {
 sub _is_vendor_consent_range_encoding {
     my $self = shift;
 
-    return is_set( $self->{data}, VENDOR_CONSENT_RANGE_ENCODING_OFFSET );
+    return
+      scalar( is_set( $self->{data}, VENDOR_CONSENT_RANGE_ENCODING_OFFSET ) );
 }
 
 sub _parse_range_section {
-    my ( $self, $vendor_bits_required, $start_bit ) = @_;
+    my ( $self, $max_id, $offset ) = @_;
 
-    my $range_section = GDPR::IAB::TCFv2::RangeSection->new(
-        data                 => $self->{data},
-        start_bit            => $start_bit,
-        vendor_bits_required => $vendor_bits_required,
-    );
+    my ( $range_section, $next_offset ) =
+      GDPR::IAB::TCFv2::RangeSection->Parse(
+        data    => $self->{data},
+        offset  => $offset,
+        max_id  => $max_id,
+        options => $self->{options},
+      );
 
-    return ( $range_section, $range_section->current_offset );
+    return ( $range_section, $next_offset );
 }
 
 sub _parse_bitfield {
-    my ( $self, $vendor_bits_required, $start_bit ) = @_;
+    my ( $self, $max_id, $offset ) = @_;
 
-    my $bitfield = GDPR::IAB::TCFv2::BitField->new(
-        data                 => $self->{data},
-        start_bit            => $start_bit,
-        vendor_bits_required => $vendor_bits_required,
+    my ( $bitfield, $next_offset ) = GDPR::IAB::TCFv2::BitField->Parse(
+        data    => $self->{data},
+        offset  => $offset,
+        max_id  => $max_id,
+        options => $self->{options},
     );
 
-    return ( $bitfield, $start_bit + $vendor_bits_required );
+    return ( $bitfield, $next_offset );
 }
 
 sub looksLikeIsConsentVersion2 {
@@ -595,7 +636,7 @@ GDPR::IAB::TCFv2 - Transparency & Consent String version 2 parser
 
 =head1 VERSION
 
-Version 0.081
+Version 0.082
 
 =head1 SYNOPSIS
 
@@ -697,7 +738,7 @@ a compact json representation. With C<verbose> we will present everything. See L
 
 C<compact> changes the json encoding. All fields that are a mapping of something to a boolean will be changed to an array
 of all elements keys where the value is true. This affects the following fields:  C<special_features_opt_in>,
-C<purposes_consent>, C<purposes_legitimate_interest>, C<vendor_consents> and C<vendor_legitimate_interests>. See L<TO_JSON> for more details.
+C<purpose/consents>, C<purpose/legitimate_interests>, C<vendor/consents> and C<vendor/legitimate_interests>. See L<TO_JSON> for more details.
 
 =item *
 
@@ -825,7 +866,7 @@ See also: L<GDPR::IAB::TCFv2::Constants::Purpose>.
 
 =head2 purpose_one_treatment
 
-CMPs can use the PublisherCC field to indicate the legal jurisdiction the publisher is under to help vendors determine whether the vendor needs consent for Purpose 1.
+CMPs can use the C<publisher_country_code> field to indicate the legal jurisdiction the publisher is under to help vendors determine whether the vendor needs consent for Purpose 1.
 
 Returns true if Purpose 1 was NOT disclosed at all.
 
@@ -937,43 +978,47 @@ With option C<convert_blessed>, the encoder will call this method.
 
     say $encoded;
 
-    # outputs:
+Outputs:
 
     {
-    "tc_string" : "COyiILmOyiILmADACHENAPCAAAAAAAAAAAAAE5QBgALgAqgD8AQACSwEygJyAAAAAA",
-    "consent_language" : "EN",
-    "purpose" : {
-        "consents" : [],
-        "legitimate_interests" : []
-    },
-    "cmp_id" : 3,
-    "purpose_one_treatment" : false,
-    "special_features_opt_in" : [],
-    "last_updated" : "2020-04-27T20:27:54.200000000Z",
-    "use_non_standard_stacks" : false,
-    "policy_version" : 2,
-    "version" : 2,
-    "is_service_specific" : false,
-    "created" : "2020-04-27T20:27:54.200000000Z",
-    "consent_screen" : 7,
-    "vendor_list_version" : 15,
-    "cmp_version" : 2,
-    "publisher_country_code" : "AA",
-    "vendor" : {
-        "consents" : [
-            23,
-            42,
-            126,
-            127,
-            128,
-            587,
-            613,
-            626
-        ],
-        "legitimate_interests" : []
-    }
+        "tc_string" : "COyiILmOyiILmADACHENAPCAAAAAAAAAAAAAE5QBgALgAqgD8AQACSwEygJyAAAAAA",
+        "consent_language" : "EN",
+        "purpose" : {
+            "consents" : [],
+            "legitimate_interests" : []
+        },
+        "cmp_id" : 3,
+        "purpose_one_treatment" : false,
+        "publisher" : {
+            "restrictions" : {}
+        },
+        "special_features_opt_in" : [],
+        "last_updated" : "2020-04-27T20:27:54.200000000Z",
+        "use_non_standard_stacks" : false,
+        "policy_version" : 2,
+        "version" : 2,
+        "is_service_specific" : false,
+        "created" : "2020-04-27T20:27:54.200000000Z",
+        "consent_screen" : 7,
+        "vendor_list_version" : 15,
+        "cmp_version" : 2,
+        "publisher_country_code" : "AA",
+        "vendor" : {
+            "consents" : [
+                23,
+                42,
+                126,
+                127,
+                128,
+                587,
+                613,
+                626
+            ],
+            "legitimate_interests" : []
+        }
     }
 
+
 If L<JSON> is installed, the C<TO_JSON> method will use C<JSON::true> and C<JSON::false> as boolean value.
 
 By default it returns a compacted format where we omit the C<false> on fields like C<vendor_consents> and we convert the dates 
