@@ -12,8 +12,12 @@ sub new {
     croak "missing field 'start'" unless defined $args{start};
     croak "missing field 'end'"   unless defined $args{end};
 
-    my $start = $args{start};
-    my $end   = $args{end};
+    croak "missing 'options'"      unless defined $args{options};
+    croak "missing 'options.json'" unless defined $args{options}->{json};
+
+    my $start   = $args{start};
+    my $end     = $args{end};
+    my $options = $args{options};
 
     croak "field 'start' should be a non zero, positive integer"
       if $start <= 0;
@@ -21,8 +25,9 @@ sub new {
     croak "ops start should not be bigger than end" if $start > $end;
 
     my $self = {
-        start => $start,
-        end   => $end,
+        start   => $start,
+        end     => $end,
+        options => $options,
     };
 
     bless $self, $klass;
@@ -36,6 +41,26 @@ sub contains {
     return $self->{start} <= $id && $id <= $self->{end};
 }
 
+sub all {
+    my $self = shift;
+
+    my ( $false, $true ) = @{ $self->{options}->{json}->{boolean_values} };
+
+    return [ $self->{start} .. $self->{end} ];
+}
+
+sub TO_JSON {
+    my $self = shift;
+
+    if ( !!$self->{options}->{json}->{compact} ) {
+        return [ $self->{start} .. $self->{end} ];
+    }
+
+    my ( $false, $true ) = @{ $self->{options}->{json}->{boolean_values} };
+
+    return { map { $_ => $true } $self->{start} .. $self->{end} };
+}
+
 1;
 __END__
 
@@ -67,3 +92,7 @@ Will die if start is bigger than end.
 Return true if the id is present on the range [start, end]
 
     my $ok = $range->contains(15);
+
+=head2 all
+
+Returns an arrayref of all vendors mapped with the bit enabled (or all vendors in range).
